import re

from jupyter_interval_widget import Interval
from ipywidgets.widgets import Button, Output, HBox, VBox
from IPython.display import display, clear_output, Javascript
from IPython.utils.py3compat import input

from .haltexecution import HaltExecution

MAGICS_REGEX = r"^((\n*(%%.+)\n)+)"

class EditorWidget(VBox):
    def __init__(self, file_manager, shell, line, cell, path, no_eval, \
                 conflict_resolution_strategy, pooling_rate=None, debug=False):
        super(EditorWidget, self).__init__()
        self.file_manager = file_manager
        self.shell = shell
        self.line = line
        self.cell = cell
        self.path = path
        self.no_eval = no_eval
        self.conflict_resolution_strategy = conflict_resolution_strategy
        self.pooling_rate =  pooling_rate if pooling_rate > 0 else None
        self.debug = debug
        self.cell_magics = re.match(MAGICS_REGEX, "%%source " + self.line + "\n" + self.cell).group(1)
        self.cell_content_to_save = re.sub(MAGICS_REGEX, "", self.cell) + "\n"
        self.status_output = Output()
        self.file_change_monitor = Interval(value=self.pooling_rate)
        self.file_change_monitor.on_tick(lambda widget: self.check_external_modification())
        self.overwrite_button = Button(description="Overwrite", button_style="danger", layout={ "display": "none" })
        self.overwrite_button.on_click(lambda widget: self.save_file_and_run_cell())
        self.refresh_button = Button(description="Refresh", button_style="danger", layout={ "display": "none" })
        self.refresh_button.on_click(lambda widget: self.refresh_cell())
        self.cell_output = Output()

        self.file_status_widgets = HBox(children=(self.status_output, self.overwrite_button, self.refresh_button))
        if self.pooling_rate:
            self.file_status_widgets.children += (self.file_change_monitor,)

        self.children = (self.file_status_widgets, self.cell_output)
        if self.debug:
            self.debug_output = Output()
            self.children = (self.debug_output,) +self.children

        self.state = "Cell was just run"

    def eval_cell(self):
        if self.no_eval:
            return
        with self.cell_output:
            clear_output(wait=True)
            self.shell.run_cell(self.cell, store_history=False)

    def save_file_and_run_cell(self):
        self.file_manager.save(self.path, self.cell_content_to_save)
        self.eval_cell()

    def refresh_cell(self):
        self.file_manager.manage(self.path, self.file_manager.read(self.path))
        js = """
            var cmd = "import os; import sys; f = open('%s', 'r'); sys.stdout.write(f.read()); f.close()";
            function handle_output(msg) {
                var cell_magic_lines = IPython.notebook.get_selected_cell().get_text().match(/^(%%%%.*\\n)*/)[0]
                var file_content = msg.content.text;
                IPython.notebook.get_selected_cell().set_text(cell_magic_lines + file_content);
            }
            var callback = {'output': handle_output};
            IPython.notebook.kernel.execute(cmd, {iopub: callback}, {silent: false});
        """
        clear_output(wait=True)
        display(self)
        display(Javascript(js % (self.path,)))


    def display_status(self, status):
        self.status_output.layout = {}
        with self.status_output:
            clear_output(wait=True)
            display(status)

    def display_debug(self, message):
        from datetime import datetime
        with self.debug_output:
            clear_output(wait=True)
            display(message + " " + str(datetime.now()))

    def show_button(self, button):
        button.layout = {}

    def prompt_conflict_resolution(self, state):
        if state == "File was deleted by another process":
            answers = {'1': 'Keep Cell Contents and Continue Execution', \
                       '2': 'Overwrite File and Continue Execution', \
                       '3': 'Abort'}
        else:
            answers = {'1': 'Keep Cell Contents and Continue Execution', \
                       '2': 'Overwrite File and Continue Execution', \
                       '3': 'Halt Execution and Refresh Cell Contents', \
                       '4': 'Abort'}

        prompt_text = state + "\n\n" + "\n".join(['%s: %s' % (key, value) for (key, value) in sorted(answers.items())]) + "\n\n"

        ans = None

        while ans not in answers.keys():
            ans = input(prompt_text + ' ')

        clear_output(wait=True)
        display(self)

        return answers[ans]

    def display_conflict_resolution(self):
        self.display_status(self.state)
        if self.state != "File was deleted by another process":
            self.show_button(self.refresh_button)
        self.show_button(self.overwrite_button)

    def process_cell_execution(self):
        state = self.file_manager.get_initial_state(self.path, self.cell_content_to_save)
        self.state = state

        if self.debug: self.display_debug(self.state)

        if self.state == "In sync, new content":
            self.save_file_and_run_cell()
        elif self.state == "In sync, unchanged content":
            self.eval_cell()
        elif self.state == "New file":
            self.save_file_and_run_cell()
        elif self.state == "File exists with same contents":
            self.file_manager.manage(self.path, self.cell_content_to_save)
            self.eval_cell()
        elif self.state in ("File exists with different contents", \
                            "File was changed by another process", \
                            "File was deleted by another process"):
            if self.conflict_resolution_strategy == "Prompt":
                resolution = self.prompt_conflict_resolution(self.state)
            else:
                resolution = self.conflict_resolution_strategy
            if resolution == "Keep Cell Contents and Continue Execution":
                self.eval_cell()
                self.display_conflict_resolution()
            elif resolution == "Overwrite File and Continue Execution":
                self.file_manager.save(self.path, self.cell_content_to_save)
                self.eval_cell()
            elif resolution == "Halt Execution and Refresh Cell Contents":
                self.shell.set_next_input(self.cell_magics + self.file_manager.read(self.path), replace=True)
                raise HaltExecution
            elif resolution == "Abort":
                self.display_conflict_resolution()
                raise HaltExecution

    def check_external_modification(self):
        state = self.file_manager.get_state_change(self.path, self.cell_content_to_save)
        if state == None or state == self.state:
            return
        self.state = state

        if self.debug: self.display_debug(self.state)

        for button in (self.overwrite_button, self.refresh_button):
            button.layout = { "display": "none" }
        self.status_output.layout = { "display": "none" }

        if state == "In sync, unchanged content":
            pass
        if state in ("File was changed by another process", \
                     "File was deleted by another process", \
                     "File exists at this path"):
            self.display_conflict_resolution()



import os
import errno

class FileManager(object):
    def __init__(self):
        super(FileManager, self).__init__()
        self.managed_files = {}

    def manage(self, path, contents):
        self.managed_files[path] = {
            "mtime" : os.path.getmtime(path),
            "digest": self.digest(contents)
        }

    def read(self, path):
        with open(path, 'r') as f:
            return f.read()

    def digest(self, contents):
        return hash(contents)

    def mkdir_p(self, path):
        if path == "" or os.path.exists(path):
            return
        os.makedirs(path)

    def file_exists(self, path):
        return os.path.exists(path)

    def same_contents(self, path, contents):
        return self.read(path) == contents

    def save(self, path, contents):
        self.mkdir_p(os.path.dirname(path))
        with open(path, "w") as file:
            file.write(contents)
        self.manage(path, contents)

    def file_is_managed(self, path):
        return path in self.managed_files.keys()

    def is_file_in_sync(self, path):
        if not self.file_is_managed(path):
            return False
        managed_file = self.managed_files[path]
        if managed_file["mtime"] == os.path.getmtime(path):
            return True
        return managed_file["digest"] == self.digest(self.read(path))

    def is_new_content(self, path, current_contents):
        managed_file = self.managed_files[path]
        return managed_file["digest"] != self.digest(current_contents)

    def is_file_in_sync_new_content(self, path, current_contents):
        return self.file_exists(path) and \
               self.is_file_in_sync(path) and \
               self.is_new_content(path, current_contents)

    def is_file_in_sync_unchanged_content(self, path, current_contents):
        return self.file_exists(path) and \
               self.is_file_in_sync(path) and \
               not self.is_new_content(path, current_contents)

    def is_new_file(self, path):
        return not self.file_is_managed(path) and \
               not self.file_exists(path)

    def is_file_exists_with_same_contents(self, path, current_contents):
        return not self.file_is_managed(path) and \
               self.file_exists(path) and \
               self.same_contents(path, current_contents)

    def is_overwrite_existing_file(self, path, current_contents):
        return not self.file_is_managed(path) and \
               self.file_exists(path) and \
               not self.same_contents(path, current_contents)

    def is_file_was_changed_by_another_process(self, path, current_contents):
        return self.file_is_managed(path) and \
               self.file_exists(path) and \
               not self.is_file_in_sync(path)

    def is_managed_file_was_deleted(self, path):
        return self.file_is_managed(path) and \
               not self.file_exists(path)

    def is_file_was_deleted(self, path):
        return not self.file_exists(path)

    def is_unmanaged_file_exists(self, path):
        return not self.file_is_managed(path) and \
               self.file_exists(path)

    def get_initial_state(self, path, current_contents):
        if self.is_file_in_sync_new_content(path, current_contents): return "In sync, new content"
        if self.is_file_in_sync_unchanged_content(path, current_contents): return "In sync, unchanged content"
        if self.is_new_file(path): return "New file"
        if self.is_file_exists_with_same_contents(path, current_contents): return "File exists with same contents"
        if self.is_overwrite_existing_file(path, current_contents): return "File exists with different contents"
        if self.is_file_was_changed_by_another_process(path, current_contents): return "File was changed by another process"
        if self.is_managed_file_was_deleted(path): return "File was deleted by another process"
        raise Exception("Unknown State")

    def get_state_change(self, path, current_contents):
        if self.is_file_in_sync_unchanged_content(path, current_contents): return "In sync, unchanged content"
        if self.is_file_was_changed_by_another_process(path, current_contents): return "File was changed by another process"
        if self.is_file_was_deleted(path): return "File was deleted by another process"
        if self.is_unmanaged_file_exists(path): return "File exists at this path"



import setuptools

setuptools.setup(
    name="jupyter-source-magic",
    version='0.1.2',
    url='https://github.com/srizzo/jupyter-source-magic',
    author="Samuel Rizzo",
    author_email='rizzolabs@gmail.com',
    description="A Jupyter Notebook %%magic to quickly edit and run source code",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    license='MIT',
    install_requires=[
        'jupyter',
        'ipywidgets',
        'jupyter-interval-widget'
    ],
    keywords=['ipython', 'jupyter'],
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.7',
    ]
)



from IPython.display import display
from IPython.core.magic_arguments import magic_arguments, argument, parse_argstring
from IPython.core.magic import line_magic, cell_magic, Magics, magics_class

from .filemanager import FileManager
from .editorwidget import EditorWidget

@magics_class
class SourceMagics(Magics):
    def __init__(self, shell):
        super(SourceMagics, self).__init__(shell)
        self.file_manager = FileManager()

    @line_magic
    @magic_arguments()
    @argument('path', type=str, help='Source path')
    def loadsource(self, line=''):
        """
        Load file for editing (replaces cell content)
        """
        args = parse_argstring(self.loadsource, line)
        self.shell.set_next_input("%%source " + line + "\n" + self.file_manager.read(args.path), replace=True)

    @cell_magic
    @magic_arguments()
    @argument('path', type=str, help='Source path')
    @argument("-p", "--pooling-rate",
              type=int,
              default=500,
              help="pooling rate for external file modifications in milliseconds. Set to 0 to disable it (default: 500)"
    )
    @argument("--no-eval",
              help="do not evaluate contents",
              action='store_true'
    )
    @argument("--debug",
              help="Display debugging info",
              action="store_true"
    )
    def source(self, line='', cell=""):
        """
        Run and write the contents of the cell to a file.

        If file exists and cell and file are out of sync, prompts what to do (see args to override).

        Periodically pools for external file modifications (see args to override).
        """
        args = parse_argstring(self.source, line)
        editor_widget = EditorWidget(file_manager = self.file_manager, \
                                     shell = self.shell, \
                                     line = line, \
                                     cell = cell, \
                                     path = args.path, \
                                     no_eval = args.no_eval, \
                                     conflict_resolution_strategy = args.conflict_resolution_strategy, \
                                     pooling_rate = args.pooling_rate, \
                                     debug = args.debug)
        display(editor_widget)
        editor_widget.process_cell_execution()
        editor_widget.check_external_modification()

    arg_group = source.parser.add_mutually_exclusive_group()
    arg_group.add_argument('--prompt', \
                           help="prompt what to do if cell and file are out of sync (default)", \
                           action='store_const', dest='conflict_resolution_strategy', const='Prompt', default='Prompt')
    arg_group.add_argument('--keep-both', \
                           help="if cell and file are out of sync keep both untouched, proceed with execution", \
                           action='store_const', dest='conflict_resolution_strategy', const='Keep Cell Contents and Continue Execution')
    arg_group.add_argument('--overwrite-existing', \
                           help="overwrite existing file without confirmation", \
                           action='store_const', dest='conflict_resolution_strategy', const='Overwrite File and Continue Execution')
    arg_group.add_argument('--abort', \
                           help="abort notebook execution if cell and file are out of sync", \
                           action='store_const', dest='conflict_resolution_strategy', const='Abort')

    source.__doc__ = source.parser.format_help()
